from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyWith, MoviepyArgument
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition
from yta_general_utils.math.rate_functions import RateFunction
from moviepy.Clip import Clip
from typing import Union


class CirclesAtPositionEffect(Effect):
    def apply(self, video: Clip, position: Union[Position, Coordinate, tuple] = Position.CENTER) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlurEffect.apply, locals(), ['video'])
        background_video = ClipGenerator.get_default_background_video(duration = video.duration)

        return self.apply_over_video(video, background_video, position)
    
    # TODO: What about this (?)
    def apply_over_video(self, video: Clip, background_video: Clip, position: Union[Position, Coordinate] = Position.CENTER) -> Clip:
        arg = MoviepyArgument(position, position, TFunctionSetPosition.linear_doing_circles, RateFunction.linear)

        return MoviepyWith().apply_over_video(video, background_video, with_position = arg)


"""
Using the 'get_frame' in moviepy returns a numpy array of 3 or 1 
dimension depending on the type of clip. If the clip is a main 
clip, a 3d array with values between [0, 255] is returned (one
[255, 255, 255] array would be a white pixel). If the clip is a
mask clip, a 1d array with values between [0, 1] is returned (the
1 completely transparent, the 0 is not).

Thats why you need to normalize or denormalize those values to
work with them because they are different and turning frames into
an image would need you to have the same range [0, 1] or [0, 255].

Check the Pillow, ImageIO and other libraries to see what kind of
numpy arrays are needed to write (or read) images.

TODO: This was in the previous 'frames.py' file that was moved to
a whole module in a folder. Do I need this explanation here (?)
"""


"""
This is a class to keep the code that allows me to
insert an image into another image that contains
a greenscreen region but that is not rectangular,
so we need to make some transformations to fit the
expected region and position.

TODO: Please, this is not fine at all and need 
review, work and love
"""
from yta_general_utils.image.converter import ImageConverter
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy.Clip import Clip
from moviepy import ImageClip, concatenate_videoclips

import cv2
import numpy as np


def _detect_image_corners_with_hsv(image_filename: str):
    """
    Detect the greenscreen corners by applying a hsv mask.
    This method should be improved as it is not detecting 
    the greenscreen properly.

    TODO: Maybe append this to the ImageRegionFinder class.
    """
    image = cv2.imread(image_filename)

    hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

    # Define the mask color in HSV values
    low = np.array([35, 50, 50])
    high = np.array([85, 255, 255])

    # Mask to detect green pixels
    mask = cv2.inRange(hsv, low, high)

    # Erode an dilate to improve the mask quality (but it doesn't)
    # work properly for sure
    # mask = cv2.erode(mask, None, iterations=2)  # Erosiona la máscara para eliminar ruido
    # mask = cv2.dilate(mask, None, iterations=2)  # Dilata la máscara para conectar áreas cercanas

    # Find region contours
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    corners = []
    # Draw green rectangle around the detected region
    # TODO: I should have only one rectangle
    for contour in contours:
        # We approach the shape to a polygon
        # Original factor was 0.02
        epsilon = 0.01 * cv2.arcLength(contour, True)
        polygon_approach = cv2.approxPolyDP(contour, epsilon, True)

        # We accept a 4 corners polygon
        if len(polygon_approach) == 4:
            corners = [(corner[0][0], corner[0][1]) for corner in polygon_approach]

    return corners

def _order_corners(corners):
    """
    This method orders the coordinates clockwise from upper
    left to lower left corner.
    """
    # We ensure they are numpy arrays to work with them
    corners = np.array(corners, dtype = 'float32')

    # We apply a 'x' order from lower to higher
    corners = corners[np.argsort(corners[:, 1])]
    upper_left, upper_right = sorted(corners[:2], key = lambda p: p[0])
    lower_left, lower_right = sorted(corners[2:], key = lambda p: p[0])

    return [upper_left, upper_right, lower_right, lower_left]

def insert_image_into_3d_greenscreen(greenscreen_filename: str, image_to_insert_filename: str, output_filename: str):
    """
    Inserts the provided 'image_to_insert_filename' in the 3d
    greenscreen region available in the also provided 
    'greenscreen_filename' and writes it as a local file with
    the given 'output_filename' file name.
    """
    corners = _detect_image_corners_with_hsv(greenscreen_filename)

    image = greenscreen_filename
    if PythonValidator.is_string(greenscreen_filename):
        image = cv2.imread(greenscreen_filename)

    image_to_inject = image_to_insert_filename
    if PythonValidator.is_string(image_to_insert_filename):
        image_to_inject = cv2.imread(image_to_insert_filename)

    # We need coordinates in clockwise order following the next
    # format: ul, ur, br, bl (u = upper, b = bottom)
    corners = _order_corners(corners)
    corners = np.array(corners, dtype = 'float32')

    # Corners of the image to inject
    alto, ancho = image_to_inject.shape[:2]
    shape_to_inject = np.array([
        [0, 0],
        [ancho - 1, 0],
        [ancho - 1, alto - 1],
        [0, alto - 1]
    ], dtype = 'float32')

    # Calculate new matrix to fit the new region
    matriz = cv2.getPerspectiveTransform(shape_to_inject, corners)

    # Transform the perspective to fit the region
    inserted_image = cv2.warpPerspective(image_to_inject, matriz, (image.shape[1], image.shape[0]), flags = cv2.INTER_NEAREST)

    # Create a mask of the inserted image (only non-transparent part)
    inserted_mask = np.zeros_like(image, dtype = np.uint8)
    cv2.fillConvexPoly(inserted_mask, corners.astype(int), (255, 255, 255))

    # Merge original image with new image applying the mask
    final_image = cv2.bitwise_and(image, cv2.bitwise_not(inserted_mask))
    final_image = cv2.add(final_image, inserted_image)

    # TODO: I need to be more accurate on detecting the corners. I
    # think I could do it manually by detecting all pixels and getting
    # firstly those that are more on the left

    # TODO: I would need to detect the corners better, and also apply
    # another kind of replacement (put the original image in the 
    # foreground, with the region as transparent pixels, and place the
    # the image to insert in the region but also adding some pixels
    # to the corners to ensure it fits the region)

    # TODO: Adapt this to return the image object and not only writing
    # it
    cv2.imwrite(output_filename, final_image)

    return output_filename

def insert_video_into_3d_greenscreen(greenscreen_filename: str, video: Clip, output_filename: str):
    """
    Inserts the provided 'video' into the also provided 3d 
    greenscreen in the file 'greenscreen_filename'. This 
    method will write the file as 'output_filename' if this
    is provided, and will return the new video anyways.
    """
    # We create each image by inserting each frame and then 
    # concatenate all of them to make a new video
    imageclips = [
        ImageClip(insert_image_into_3d_greenscreen(greenscreen_filename, ImageConverter.numpy_image_to_opencv(frame), 'a.png'), duration = 1 / video.fps).with_fps(video.fps) 
        for frame in video.iter_frames()
    ]

    video = concatenate_videoclips(imageclips)

    if (output_filename):
        video.write_videofile(output_filename)

    return video


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.programming.parameter_validator import NumberValidator
from moviepy.video.fx import MultiplySpeed
from moviepy import Clip


class FitDurationEffect(Effect):
    """
    This effect changes the speed of the video to fit
    the requested 'duration', that will accelerate
    or decelerate the video speed.
    """
    def apply_over_video(self):
        # TODO: What do I do with this (?)
        pass

    def apply(self, video: Clip, duration: float = None) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlurEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        if not NumberValidator.is_number_between(duration, 0, 120, False):
            raise Exception(f'The provided "duration" parameter "{str(duration)}" is not a valid number between 0 (not included) and 120 (included).')

        return MultiplySpeed(final_duration = duration).apply(video)
    
# TODO: Maybe I can create another effect to set
# the speed by multiplying it by a factor, but 
# by now this one is enough


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionResize
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyWithPrecalculated
from yta_multimedia.video import MPVideo
from yta_multimedia.video.position import Position
from yta_general_utils.math.rate_functions import RateFunction
from moviepy.Clip import Clip


class ZoomInDisplayAndZoomOutEffect(Effect):
    """
    Makes the provided clip have a zoom in effect, then
    being displayed at all, and later being zoomed out
    before finishing.

    This effect is recommended to be used with a pure 
    white background.
    """
    def apply(self, video: Clip) -> Clip:
        background_video = ClipGenerator.get_default_background_video(duration = video.duration)

        return self.apply_over_video(video, background_video)
    
    def apply_over_video(self, video: Clip, background_video: Clip):
        video = VideoParser.to_moviepy(video)
        background_video = VideoParser.to_moviepy(background_video)

        video_handler = MPVideo(video)
        zoom_duration = 30 * video_handler.frame_duration

        # TODO: This has to be passed as argument, but it is a good
        # background for the movement
        #first_white_background = ClipGenerator.generate_color_background((1920, 1080), [255, 255, 255], video.duration, video.fps)
        
        resizes = []
        for t in video_handler.frames_time_moments:
            if t < zoom_duration:
                resizes.append(TFunctionResize.resize_from_to(t, zoom_duration, 1.0, 0.7, RateFunction.ease_in_expo))
            elif t < (video.duration - zoom_duration):
                resizes.append(0.7)
            else:
                resizes.append(TFunctionResize.resize_from_to(t - (video.duration - zoom_duration), zoom_duration, 0.7, 1.0, RateFunction.ease_in_expo))

        positions = [Position.CENTER.get_moviepy_upper_left_corner_tuple((video.w * resizes[i], video.h * resizes[i]), background_video.size) for i in range(len(resizes))]

        return MoviepyWithPrecalculated().apply_over_video(video, background_video, resized_list = resizes, with_position_list = positions)



from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE
from yta_multimedia.video.generation.manim.utils.dimensions import fitting_text
from yta_multimedia.video.position import Position
from manim import *
from typing import Union


class RainOfWordsManimAnimationWrapper(BaseManimAnimationWrapper):
    """
    This is a rain of the provided 'words' over the screen, that
    appear in random positions.
    """
    words: list[str] = None
    duration: float = None

    def __init__(self, words: list[str], duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if not PythonValidator.is_list_of_string(words):
            exception_messages.append('The "words" parameter provided is not a list of strings.')
        
        if not NumberValidator.is_positive_number(duration) or not NumberValidator.is_number_between(duration, 0, 100, do_include_lower_limit = False):
            exception_messages.append('The "duration" parameter provided is not a positive number between (0, 100] (zero is not valid).')

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        self.words = words
        self.duration = duration
        super().__init__(RainOfWordsManimAnimationGenerator)

class RainOfWordsManimAnimationGenerator(BaseManimAnimation):
    """
    This is a rain of the provided 'words' over the screen, that
    appear in random positions.
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        each_word_time = self.parameters['duration'] / len(self.parameters['words'])
        # Adjust the divisor number to modify word size
        for word in self.parameters['words']:
            text = fitting_text(word, MOVIEPY_SCENE_DEFAULT_SIZE[0] / 6)
            # TODO: This was previously considering the limits to
            # make the text be always inside the scene, but now...
            #random_coords = Position.RANDOM_INSIDE.get_manim_position_center((text.width, text.height))
            random_coords = Position.RANDOM_INSIDE.get_manim_position_center()
            text.move_to(random_coords)
            self.add(text)
            self.wait(each_word_time)


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper
from yta_multimedia.resources import Resource
from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_general_utils.text.transformer import remove_accents
from manim import *
from typing import Union


class MagazineTextStaticManimAnimationWrapper(BaseManimAnimationWrapper):
    """
    Makes an animation in which the provided 'text' appears
    with special magazine characters. The text is shown
    without any animation. It is only one row of text 
    limited to 30 characters.
    """
    text: str = None
    duration: float = None

    def __init__(self, text: str, duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if not text:
            exception_messages.append('No "text" parameter provided.')
        
        if not NumberValidator.is_positive_number(duration) or not NumberValidator.is_number_between(duration, 0, 100, do_include_lower_limit = False):
            exception_messages.append('The "duration" parameter provided is not a positive number between (0, 100] (zero is not valid).')

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        self.text = text
        self.duration = duration
        super().__init__(MagazineTextStaticManimAnimationGenerator)

class MagazineTextStaticManimAnimationGenerator(BaseManimAnimation):
    """
    Makes an animation in which the provided 'text' appears
    with special magazine characters. The text is shown
    without any animation. It is only one row of text 
    limited to 30 characters.
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        DISTANCE_BETWEEN_LETTERS = 0.30

        text = remove_accents(self.parameters['text'])

        elements = []
        # TODO: Make the text fit attending to how long it is
        for char in text:
            elements_len = len(elements)
            new_letter = self.__get(char)

            if elements_len > 0:
                new_letter.move_to(elements[elements_len - 1]).shift(RIGHT * DISTANCE_BETWEEN_LETTERS)

            elements.append(new_letter)

        all_mobjects = Group(*elements)
        all_mobjects.move_to((0, 0, 0)) # We place all elements in the center

        # We need to make the animation fit the 
        self.add(*all_mobjects)
        self.wait(self.parameters['duration'])

    # TODO: This below is repeated in the other magazine scene
    def __get(self, letter: str = 'a'):
        if not letter:
            return None
        
        letter = letter[0:1].lower()
        mobject_group = None

        if letter == 'a':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1p_NrqZEf3lhQl_lB_PMwgI4FNIbVkSs-/view?usp=sharing', 0.2, 0)
        elif letter == 'b':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1KO0WsYNcocowFeG7Uql_71dn4tlEyUEJ/view?usp=sharing', 0.2, 0)
        elif letter == 'c':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1pFC5l3C0JylftXsDajI2hYJwgFgHFHeX/view?usp=sharing', 0.2, 0)
        elif letter == 'd':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1f7MPrRmczO-oaQYG5tdadg0-q7X8Avm9/view?usp=sharing', 0.2, 0)
        elif letter == 'e':
            mobject_group = self.__get_letter(letter, '#325e2e', 'https://drive.google.com/file/d/1JLb2dPK5AdGqk3MdF1KzKwQnisU1fYwf/view?usp=sharing', 0.2, 0)
        elif letter == 'f':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1QA2u8ppuZ9nWFlbfyOBDP3sZ0GRRAcRD/view?usp=sharing', 0.2, 0)
        elif letter == 'g':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1-3PlY9N-X83gyYQWX-M-t0R_gUkp8TMp/view?usp=sharing', 0.2, 0)
        elif letter == 'h':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1_e5SzuadQUzMia_KhiLa2p9qntxg-kIC/view?usp=sharing', 0.2, 0)
        elif letter == 'i':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1pWGKVyDGyXcLX2Xwg8pxAV0FHSchB9_J/view?usp=sharing', 0.2, 0)
        elif letter == 'j':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/17fLo90TghQIRsrLvIm7gRIwKjJkcNgun/view?usp=sharing', 0.2, 0)
        elif letter == 'k':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1yoLkzH1o8vZ4RjVCDWza6WCWvZ8y5MxN/view?usp=sharing', 0.2, 0)
        elif letter == 'l':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1cZAncrpB0j8FD64opEzQdyeCtQVJbajU/view?usp=sharing', 0.2, 0)
        elif letter == 'm':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1XQ02p-veaiSWLyx2PozCfK2QhflDCnpu/view?usp=sharing', 0.2, 0)
        elif letter == 'n':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1gp-9ZLAVVPeOO5uitF6zVOJqeK6443R8/view?usp=sharing', 0.2, 0)
        # TODO: Make 'ñ'
        elif letter == 'o':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/13H8nmkH9UiTn7vG57WgpC0v5S_UNATPM/view?usp=sharing', 0.2, 0)
        elif letter == 'p':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1YNSFeSAujWWFuwrqU_A-iw_Xk-l2vXvU/view?usp=sharing', 0.2, 0)
        elif letter == 'q':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1k4MuXPS9M2H4Z3BCnCf64yXcBwf6QEKO/view?usp=sharing', 0.2, 0)
        elif letter == 'r':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1nM_jNdzGGV60Z8fAjJp-Q43RXCmwsqh0/view?usp=sharing', 0.2, 0)
        elif letter == 's':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1bJC8IAMsHOdqeVxdPBhuV3_KYIE0DZSZ/view?usp=sharing', 0.2, 0)
        elif letter == 't':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1H4t4iS_yrsAqg7QEdAhxwDOOahBh54dl/view?usp=sharing', 0.2, 0)
        elif letter == 'u':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1M_uMgx9CBkwM0tKTMH_YiyGRiYPIxbXa/view?usp=sharing', 0.2, 0)
        elif letter == 'v':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/12SGuNyNdF03ua2Ro7GIoAf7ETUB0jDWl/view?usp=sharing', 0.2, 0)
        elif letter == 'w':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1aDQ0vhhrnh-BzVgtnTv9HOFSm7cgxycc/view?usp=sharing', 0.2, 0)
        elif letter == 'x':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1wjCkyibwEYz3G19XZiO9viGkNhFRgY0M/view?usp=sharing', 0.2, 0)
        elif letter == 'y':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1aIEu0EP8a-tDHY6aGvAnFGtGyFBn9M7D/view?usp=sharing', 0.2, 0)
        elif letter == 'z':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1k7EryiaWPX8HRlnB7tJtEL1eI85z7KD_/view?usp=sharing', 0.2, 0)
        # Special ones
        elif letter == '¿': # Background is the same as '?'
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1cGm0r4cdDi4b4ZSS_VxH3yJdyJ5sGB6Q/view?usp=sharing', 0.2, 0)
        elif letter == '?': # Background is the same as '¿'
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1cGm0r4cdDi4b4ZSS_VxH3yJdyJ5sGB6Q/view?usp=sharing', 0.2, 0)
        elif letter == '¡': # Background is the same as '!'
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1J4aL2yf1V_lCjbxzgXE1HyoXoyBRVXwx/view?usp=sharing', 0.2, 0)
        elif letter == '!': # Backgroudn is the same '¡'
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1J4aL2yf1V_lCjbxzgXE1HyoXoyBRVXwx/view?usp=sharing', 0.2, 0)
        # General ones (those who I don't specify manually)
        # elif letter == ' ':
        #     mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1fLol_Fe_AiQ6ZDY8pWW6yBXSFg9PnTb_/view?usp=sharing', 0.2, 0)
        else:   # Any other sign, even the space
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1fLol_Fe_AiQ6ZDY8pWW6yBXSFg9PnTb_/view?usp=sharing', 0.2, 0)

        return mobject_group

    def __get_letter(self, letter: str = 'a', color: ManimColor = WHITE, google_drive_url: str = None, background_scale: float = 0.2, background_shift: float = 0):
        """
        Generates the default Mobject Group for our own characters
        """
        if not letter:
            return None

        if not google_drive_url:
            return None

        letter = letter[0:1]

        if letter == ' ':
            return Group(Text('a', font_size = 48, color = color).set_opacity(0))
        
        letter_background_filename = letter
        if letter == '¿':
            letter_background_filename = 'open_question_mark'
        elif letter == '?':
            letter_background_filename = 'close_question_mark'
        elif letter == '¡':
            letter_background_filename = 'open_exclamation_mark'
        elif letter == '!':
            letter_background_filename = 'close_exclamation_mark'

        TMP_FILENAME = Resource.get(google_drive_url, 'resources/manim/magazine_letters/' + letter_background_filename + '.svg')
        letter_background = SVGMobject(TMP_FILENAME).scale(0.2).shift(UP * background_shift)
        letter_text = Text(letter, font_size = 32, color = color)
        letter_group = Group(letter_background, letter_text)
        
        return letter_group


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper
from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE
from yta_multimedia.video.generation.manim.utils.dimensions import fitting_text
from yta_general_utils.programming.parameter_validator import NumberValidator
from manim import *
from typing import Union


class TextTripletsManimAnimationWrapper(BaseManimAnimationWrapper):
    """
    The provided 'text' is splitted in triplets and appear on the screen. This animation
    lasts 'duration' seconds. Each triplet appear each 'duration' / len(words).
    """
    text: str = None
    duration: float = None

    def __init__(self, text: str, duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if not text:
            exception_messages.append('No "text" parameter provided.')
        
        if not NumberValidator.is_positive_number(duration) or not NumberValidator.is_number_between(duration, 0, 100, do_include_lower_limit = False):
            exception_messages.append('The "duration" parameter provided is not a positive number between (0, 100] (zero is not valid).')

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        self.text = text
        self.duration = duration
        super().__init__(TextTripletsManimAnimationGenerator)

class TextTripletsManimAnimationGenerator(BaseManimAnimation):
    """
    The provided 'text' is splitted in triplets and appear on the screen. This animation
    lasts 'duration' seconds. Each triplet appear each 'duration' / len(words).
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        words = self.parameters['text'].split(' ')
        # We need to adjust the array to contain a multiple of 3 number of elements
        leftover_numbers = words[len(words) - len(words) % 3:]
        if len(leftover_numbers) > 0:
            words = words[:len(words) - len(leftover_numbers)]

        words_triplets = []
        subarray = []
        for word in words:
            subarray.append(word)
            if len(subarray) == 3:
                words_triplets.append(subarray)
                subarray = []
        if leftover_numbers:
            words_triplets += [leftover_numbers]
        each_triplet_time = self.parameters['duration'] / len(words_triplets)

        for triplet in words_triplets:
            str = ' '.join(triplet)
            # I don't know how to show one word before the other
            # Helping information:
            # For example say you would like to not render and skip the begin of a video , you put self.next_section(skip_animations=True) in the line after def construct(self): and put self.next_section() before the first line of the animation you want to render.
            # Thank you: https://www.reddit.com/r/manim/comments/tq1ii8/comment/ihzogki/?utm_source=share&utm_medium=web3x&utm_name=web3xcss&utm_term=1&utm_content=share_button
            text = fitting_text(str, MOVIEPY_SCENE_DEFAULT_SIZE[0] / 2)
            # Create 3 similar texts with each word
            self.add(text)
            self.wait(each_triplet_time)
            self.remove(text)


    


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyArgument, MoviepyWith
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionResize
from yta_multimedia.video import MPVideo
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_general_utils.math.rate_functions import RateFunction
from moviepy.Clip import Clip

import numpy as np


# TODO: The use of this method must be by using the
# new MoviepyWith, not VideoEffect that will be
# removed in a near future version.
# TODO: Rename it, as it is making zoom but in an
# specific an static position, so it should be
# something like ZoomInPlaceEffect
class ZoomVideoEffect(Effect):
    """
    Creates a Zoom effect in the provided video.
    """
    def apply(cls, video: Clip, zoom_start: float, zoom_end: float, rate_func: type = RateFunction.linear):
        """
        Apply the effect on the provided 'video'.

        :param float zoom_start: The zoom at the start of the video, where a 1 is no zoom, a 0.8 is a zoom out of 20% and 1.1 is a zoom in of a 10%.

        :param float zoom_end: The zoom at the end of the video, where a 1 is no zoom, a 0.8 is a zoom out of 20% and 1.1 is a zoom in of a 10%.

        :param type rate_func: The rate function to apply in the animation effect. Must be one of the methods available in the RateFunction class.
        """
        # TODO: I have 2 ways of making a zoom. One is placing the clip
        # over a black (transparent) background and resizing in in the
        # center. The other one is resizing it and recalculating the
        # position for each frame to be at the same position but still
        # being resized

        # 1st. Use black background
        arg = MoviepyArgument(zoom_start, zoom_end, TFunctionResize.resize_from_to, rate_func)
        background_video = ClipGenerator.get_default_background_video(duration = video.duration)

        return MoviepyWith.apply_over_video(video, background_video, resized = arg)
    
        video_handler = MPVideo(video)

        # 2nd. Resize but keeping the position
        resizes = [TFunctionResize.resize_from_to(t, video.duration, zoom_start, zoom_end, rate_func) for t in video_handler.frame_time_moments]

    # TODO: 'apply_over_video' here (?)


"""
Extracted from official documentation:
- https://zulko.github.io/moviepy/getting_started/videoclips.html?highlight=mask#mask-clips

The fundamental difference between masks and standard clips is that standard clips output frames with 3 components (R-G-B) per pixel, comprised between 0 and 255, while a mask has just one composant per pixel, between 0 and 1 (1 indicating a fully visible pixel and 0 a transparent pixel). Seen otherwise, a mask is always in greyscale.
"""
from yta_multimedia.video import MPVideo
from yta_multimedia.video.parser import VideoParser
from moviepy import CompositeVideoClip
from moviepy.Clip import Clip
from typing import Union


# TODO: I think this will be deleted in the future, when refactored
# and when we confirm that here is another way of handling this
def apply_inverted_mask(video: Clip, mask_video: Clip, output_filename: Union[str, None] = None):
    """
    Applies the provided 'mask_video' with its mask inverted
    over the also provided 'video'. This is useful to make
    artistic effects. This methods applies the 
    'invert_video_mask' method to the provided 'mask_video'.
    """
    video = VideoParser.to_moviepy(video)
    # TODO: Do I need the mask, it is a mask clip not a clip
    # with a mask, so I think I don't...
    mask_video = VideoParser.to_moviepy(mask_video, True)

    mask_video = MPVideo(mask_video).invert()
    # TODO: Handle durations
    final_clip = CompositeVideoClip([video, mask_video.with_subclip(0, video.duration)], use_bgclip = True)
    final_clip = final_clip.with_audio(video.audio)

    if output_filename:
        final_clip.write_videofile(output_filename)

    return final_clip


import os

# This test file will create some audio files during the process
TEST_FOLDER = os.getcwd().replace('\\', '/') + '/yta_voice_module/test/'
TEST_TEXT = text = 'Esto es un test con el que quiero que se genere un audio narrado'


def test_coqui_tts():
    from yta_voice_module.tts.coqui import narrate
    
    narrate(TEST_TEXT, output_filename = TEST_FOLDER + 'test_coqui_tts.wav')

def test_imitate_coqui_tts():
    # TODO: Move this to method and refresh
    from TTS.api import TTS

    from TTS.api import TTS

    tts = TTS("tts_models/multilingual/multi-dataset/xtts_v2")
    #tts = TTS('xtts')

    """
    # using a specific version
    # 👀 see the branch names for versions on https://huggingface.co/coqui/XTTS-v2/tree/main
    # ❗some versions might be incompatible with the API
    tts = TTS("xtts_v2.0.2", gpu=True)

    # getting the latest XTTS_v2
    tts = TTS("xtts", gpu=True)
    """

    # input_filename can be an array of wav files
    # generate speech by cloning a voice using default settings
    tts.tts_to_file(text = 'Quiero que te relajes y que por un momento pienses que la vida es bonita. Que no merece la pena enfadarse tanto. La vida es agradecida si tú también lo eres.', file_path = TEST_FOLDER + 'test_imitate_voice_coqui_tts.wav', speaker_wav = TEST_FOLDER + 'resources/narracion_irene_albacete.mp3', language = 'es')

def test_google_tts():
    from yta_voice_module.tts.google import narrate

    narrate(TEST_TEXT, output_filename = TEST_FOLDER + 'test_google_tts.wav')

def test_microsoft_tts():
    from yta_voice_module.tts.microsoft import narrate

    narrate(TEST_TEXT, output_filename = TEST_FOLDER + 'test_microsoft_tts.wav')

def test_open_voice_tts():
    from yta_voice_module.tts.open_voice import narrate

    narrate(TEST_TEXT, output_filename = TEST_FOLDER + 'test_open_voice_tts.wav')

def test_imitate_open_voice_tts():
    from yta_voice_module.tts.open_voice import imitate_voice

    # This audio was extracted from the first 2:03 minutes of https://www.youtube.com/watch?v=ZD2h5-qAXww
    # video, that belongs to Irene Albacete's Youtube Channel (awesome voice)
    imitate_voice(TEST_TEXT, TEST_FOLDER + 'resources/narracion_irene_albacete_recortado.mp3', output_filename = TEST_FOLDER + 'test_imitate_open_voice_tts.wav')

def test_tortoise_tts():
    from yta_voice_module.tts.tortoise import narrate

    narrate(TEST_TEXT, output_filename = TEST_FOLDER + 'test_tortoise_tts.wav')

def tests():
    import os, glob

    print('Executing all tests')

    # test_coqui_tts()
    # test_google_tts()
    # test_microsoft_tts()
    # test_open_voice_tts()
    # test_imitate_open_voice_tts()
    #test_tortoise_tts()
    test_imitate_coqui_tts()

    print('Tests executed')
    
    # Set this to False to preserve test files to be able to manually check them
    do_remove = False
    if do_remove:
        for filename in glob.glob(TEST_FOLDER + '*.wav'):
            os.remove(filename)

tests()


from yta_multimedia.experimental.text.gemini_ai import GeminiAI


class AIPromptHandler:
    def __init__(self):
        self.gemini = GeminiAI()

    def generate_youtube_video_title_from_idea(self, idea: str):
        """
        Generates a Youtube video title from a prompt idea. This
        method will return the title as a string, nothing else.
        """
        # TODO: Clean, remove quotes, etc.
        prompt = 'Quiero crear un vídeo que hable sobre "' + idea + '". Necesito que generes un título que sea llamativo para que yo pueda utilizarlo en Youtube para este vídeo. El título debe atraer la atención del público, y debe tener como máximo 60 caracteres. Solo quiero que me respondas con una opción, y que tu respuesta sea única y exclusívamente el título del vídeo. No me digas que aquí tienes alguna opción ni nada por el estilo. Solo el título. Devuélveme solo la frase con el título.'

        response = self.gemini.ask(prompt)
        # TODO: Remove '*' from the generated text

        return response
    
    def generate_youtube_video_description_from_idea(self, idea: str):
        """
        Generates a Youtube video description from a prompt 
        idea. This method will return the title as a string,
        nothing else.
        """
        # TODO: Clean, remove quotes, etc.
        prompt = 'Quiero crear un vídeo que hable sobre la idea "' + idea + '". Necesito que generes una descripción que sea llamativa para que pueda utilizarlo en Youtube para ese vídeo. La descripción debe utilizar palabras clave relacionadas con la idea anterior, debe tener como máximo 300 o 400 caracteres, y debe estar dirigido hacia, primero, explicar un poco de lo que se habla en el vídeo sin decir exactamente lo que ocurre, creando hype, y segundo, estar potenciada y orientada hacia el SEO para que salga como resultados de búsqueda. Solo quiero que me respondas con una opción, y que tu respuesta sea única y exclusívamente la descripción del vídeo. No me digas que aquí tienes alguna opción ni nada por el estilo. Solo la descripción, de entre 300 y 400 caracteres. Devuélveme solo la descripción.'

        response = self.gemini.ask(prompt)
        # TODO: Remove '*' from the generated text

        return response

    def summarize_text(self, text: str, number_of_chars: int = None):
        """
        This method will summarize the provided 'text' to a new one
        with the also provided 'number_of_char' characters. If not
        provided, it will use the 20% of the 'text' characters
        amount. It returns the summarized text.

        Depending on the text, the minimum 'number_of_chars' should
        be 100 or a 10% of the provided text. Shorter texts could 
        inaccurate. The system is not able to guarantee the 
        'number_of_chars' amount of characters. It could be a huge
        difference between the desired one and the returned by this
        method.
        """
        # TODO: Clean, remove quotes, etc.
        if not number_of_chars:
            number_of_chars = 0.2 * len(text)
        number_of_chars = int(number_of_chars)

        prompt = 'Quiero que me resumas el siguiente texto: "' + text + '". Quiero que mantengas correctamente la idea inicial pero que el texto que me des esté resumido. El resumen que hagas debe contener específicamente ' + str(number_of_chars) + ' caracteres. Solo quiero que me respondas con una opción, y que tu respuesta sea única y exclusivamente el texto ya resumido. No me digas que aquí tienes alguna opción ni nada por el estilo. Solo el texto resumido, de ' + str(number_of_chars) + ' caracteres. Devuélveme solo el texto ya resumido.'

        response = self.gemini.ask(prompt)
        # TODO: Remove '*' from the generated text

        return response

    def rewrite_text(self, text: str):
        # TODO: More checkings (?)
        if not text:
            raise Exception('No "text" provided.')

        prompt = 'Eres un experto en reescribir textos. Obtienes un texto y eres capaz de convertirlo a otro que mantiene la esencia, que transmite la idea, que respeta la estructura pero que está escrito de una forma completamente distinta aunque la información que se dice es la misma. Quiero que reescribas el texto que te voy a pasar. Quiero que respetes la idea y el estilo del texto, pero que lo reescribas con el mismo número de palabras si es posible, pero con palabras distintas para que no se sepa cómo era el texto original. Solo quiero que me respondas con una opción, y que tu respuesta sea única y exclusivamente el texto ya reescrito. No me digas que aquí tienes alguna opción ni nada por el estilo. Solo el texto reescrito, de más o menos el mismo número de palabras y caracteres. Devuélveme solo el texto reescrito. El texto a reescribir es el siguiente: "' + text + '".'

        response = self.gemini.ask(prompt)
        # TODO: Remove '*' from the generated text

        return response

    def idea_to_text(self, idea: str, number_of_chars: int):
        """
        This method generates a complete text from the provided 'idea'.
        """
        # TODO: Clean the idea text
        if not number_of_chars:
            number_of_chars = 4000

        prompt = 'Eres un experto en redacción de guiones para narrar en vídeos de Youtube. A partir de una idea eres capaz de generar un texto súper completo y largo para poder narrar en el vídeo de Youtube y que la gente pueda disfrutarlo. Vas a recibir una idea y vas a generar todo el texto para ser narrado. Eres capaz de, primero, dar una introducción a los espectadores para que entiendan de qué se va a hablar. Después, entras a fondo en el tema y profundizas, dando detalles muy importantes e interesantes sobre la idea que se te va a pasar, y por último concluyes con un desenlace increíble para poner la guinda al guion del vídeo. Solo quiero que me respondas con una opción, y que tu respuesta sea única y exclusivamente el texto ya escrito. No me digas que aquí tienes alguna opción ni nada por el estilo. Solo el texto que has generado, sin ser repetitivo y con ' + str(number_of_chars) + ' caracteres de longitud. Devuélveme solo el texto. La idea de la que quiero que generes el texto es la siguiente: "' + idea + '".'

        response = self.gemini.ask(prompt)
        # TODO: Remove '*' from the generated text

        return response


from setuptools import setup, find_packages


VERSION = '0.1.88'
DESCRIPTION = 'Youtube Autónomo Multimedia Module is here.'
LONG_DESCRIPTION = 'These are all the multimedia utils we need in the Youtube Autónomo project to work in a better way. This module includes audio, image and video generation and editing utilities.'

setup(
        name = "yta_multimedia", 
        version = VERSION,
        author = "Daniel Alcalá",
        author_email = "<danielalcalavalera@gmail.com>",
        description = DESCRIPTION,
        long_description = LONG_DESCRIPTION,
        packages = find_packages(),
        install_requires = [
            # Audio module
            'pyttsx3',
            'gtts',
            'pydub',
            'pedalboard',
            'spleeter',
            'bezier',
            # Image module
            'pillow',
            'imutils'
            # Video module
            #'moviepy',
            # This below breaks everything as there is a problem with the
            # 'click' library and I have to manually reinstall 'manim' any
            # time I install 'yta_multimedia' by pip.
            #'manim' 
        ],
        
        keywords = [
            'youtube autonomo multimedia utils module'
        ],
        classifiers = [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ]
)


"""
When we need to use videos generated with manim
we have many different types of videos, and we
need to ensure that the provided wrapper class
is one of the types the method we are using is
expecting.

If we are trying to overlaying a text which is
generated with a text manim wrapper class, we
need to raise an exception if the provided class
is not a text manim wrapper class, because the
process will fail as the video generated will be
different as the expected.

All the classes we have that belong to manim video
creation have the same structure, having a wrapper
class that internally uses a generator class to
actually build the video animation, so we need
those wrapper class names. But also, the wrapper
class name is the same as the file name but in
camel case and ending in 'Wrapper'.
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper
from yta_multimedia.video.edition.resize import resize_video
from yta_multimedia.video.edition.duration import set_video_duration, ExtendVideoMode
from yta_multimedia.video.combine import VideoCombinatorAudioMode, VideoAudioCombinator
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.file.handler import FileSearchOption, FileHandler
from yta_general_utils.programming.var import snake_case_to_upper_camel_case
from yta_general_utils.programming.path import get_project_abspath
from moviepy.Clip import Clip
from moviepy import CompositeVideoClip


# TODO: Check yta_multimedia\video\combine.py because
# its functionality could end being part of this 
# VideoEditor


# TODO: Please, rename this class as this name is
# not a proper name
class VideoClassifier:
    @staticmethod
    def get_manim_wrapper_class_names_from_files(abspath: str, files_to_ignore: list[str] = []):
        """
        Obtain a list with the manim wrapper class names of
        all the available files that are in the provided
        'abspath', excluding the ones in the also given
        'files_to_ignore'. The file name is turned into the
        wrapper class name and returned.
        """
        files_to_ignore = [files_to_ignore] if PythonValidator.is_string(files_to_ignore) else files_to_ignore

        if not PythonValidator.is_list_of_string(files_to_ignore):
            raise Exception('The "files_to_ignore" parameter provided is not a valid list of strings.')

        # Transform the file name in the wrapper class that is inside
        transform_function = lambda file: snake_case_to_upper_camel_case(file.split("/")[-1].replace(".py", ""))

        return [
            f'{transform_function(file)}Wrapper'
            for file in FileHandler.get_list(abspath, FileSearchOption.FILES_ONLY, '*.py')
            if not any(file.endswith(file_to_ignore) for file_to_ignore in files_to_ignore)
        ]

    @staticmethod
    def text_manim_premades():
        """
        Get a list containing the manim text animation wrapper
        class names that can be used when text manim videos
        are needed.
        """
        #from yta_multimedia.video.generation.manim.classes.text import magazine_text_is_written_manim_animation

        return [
            'MagazineTextIsWrittenManimAnimationWrapper',
            'MagazineTextStaticManimAnimationWrapper',
            'RainOfWordsManimAnimationWrapper',
            'SimpleTextManimAnimationWrapper',
            'TestTextManimAnimationWrapper',
            'TextTripletsManimAnimationWrapper',
            'TextWordByWordManimAnimationWrapper'
        ]

        # TODO: Maybe try another way of getting all the classes
        # within a module, not a file, and identify like I tried
        # with this 'get_manim_wrapper_class_names_from_files'
        # method that is not working because files change when
        # imported as library
        return VideoClassifier.get_manim_wrapper_class_names_from_files(
            f'{get_project_abspath()}/video/generation/manim/classes/text/',
            ['__init__.py']
        )


class VideoEditor:
    _video: Clip = None

    @property
    def video(self):
        return self._video

    def __init__(self, video: Clip):
        self._video = VideoParser.to_moviepy(video, do_include_mask = True, do_calculate_real_duration = True)

    def overlay_text(self, text_generator_wrapping_instance: BaseManimAnimationWrapper):
        # TODO: The instance must be an instance of an
        # specific class (we didn't create yet) to identify
        # it as a manim text animation video generator
        # wrapping class. Maybe identify it as a text that
        # is goin to be overlayed, that can be different
        # from a text that will be the whole scene (imagine
        # a title over a white background vs a text that
        # suddenly appears over what is being shown)
        # 
        # This class contains the needed parameters (with
        # their values actually set) and the animation
        # generator class that must be called with those
        # parameters to generate the animation video.
        if not PythonValidator.is_subclass(text_generator_wrapping_instance, BaseManimAnimationWrapper) or not PythonValidator.is_an_instance(text_generator_wrapping_instance):
            raise Exception('The "text_generator_wrapping_instance" is not a valid instance of a subclass of BaseManimAnimationWrapper class.')
        
        # We validate that the provided wrapper class is 
        # about text
        if not PythonValidator.is_instance(text_generator_wrapping_instance, VideoClassifier.text_manim_premades()):
            raise Exception('The provided "text_generator_wrapping_instance" is not an instance of a manim text generation class.')
        
        video = VideoParser.to_moviepy(text_generator_wrapping_instance.generate(), do_include_mask = True)
        video = _prepare_video(self.video, video, 1)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = VideoCombinatorAudioMode.ONLY_MAIN_CLIP_AUDIO)

        return video

        # TODO: Remove this below when the above is working
        return CompositeVideoClip([
            self.video,
            VideoParser.to_moviepy(text_generator_wrapping_instance.generate(), do_include_mask = True)
        ])
    
    def overlay_video_without_alpha_fullscreen(self, video: Clip, audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
        """
        Useful to show a stock video while the main clip is
        still speaking, or to focus on the stock video.
        """
        video = VideoParser.to_moviepy(video)
        audio_mode = VideoCombinatorAudioMode.to_enum(audio_mode)

        video = _prepare_video(self.video, video, 1)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = audio_mode)

        return video

    def overlay_video_without_alpha_non_fullscreen(self, video: Clip, audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
        """
        Useful to add a video like a reel or stock while the
        main clip is still visible.
        """
        video = VideoParser.to_moviepy(video)
        audio_mode = VideoCombinatorAudioMode.to_enum(audio_mode)

        video = _prepare_video(self.video, video, 8)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = audio_mode)

        return video
    
    def overlay_video_with_alpha_fullscreen(self, video: Clip, audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
        """
        Useful to add an alphascreen, a transition or
        another kind of videos.
        """
        video = VideoParser.to_moviepy(video, do_include_mask = True)
        audio_mode = VideoCombinatorAudioMode.to_enum(audio_mode)

        video = _prepare_video(self.video, video, 1)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = audio_mode)

        return video
    
    def overlay_video_with_alpha_non_fullscreen(self, video: Clip, audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
        """
        Useful for something that I don't know right now.

        TODO: Please, improve this doc... omg
        """
        video = VideoParser.to_moviepy(video, do_include_mask = True)
        audio_mode = VideoCombinatorAudioMode.to_enum(audio_mode)

        video = _prepare_video(self.video, video, 8)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = audio_mode)

        return video
    

def _prepare_video(main_video: Clip, video: Clip, size_factor: float = 1.0):
    """
    Resize the 'video' according to the 'main_video' dimensions
    and enshort the 'video' if larger than the 'main_video'.
    """
    # We resize the 'video' to fit expected size
    video = resize_video(video, main_video.size / size_factor)
    # We ensure the video is not larger than the main one
    video = set_video_duration(video, main_video.duration, extend_mode = ExtendVideoMode.DONT_ENLARGE)

    return video

def _overlay_video(main_video: Clip, video: Clip, position: tuple = ('center', 'center'), audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
    return CompositeVideoClip([
        main_video,
        # TODO: Centered? I think this must be customizable
        video.with_position(position)
    ]).with_audio(VideoAudioCombinator(audio_mode).process_audio(main_video, video))



# 1. Videos must be 60fps both of them to simplify
# 2. The main video (background_video) must be 1920x1080 always,
#    and the other ones must be 1920x1080 or smaller
# 3. Duration of the video cannot be larger than the main video

# We should add a VideoModifications matrix in which we have
# layers that indicate the moment in which the modification has
# to be applied. Layer 1 will be prior, so once all layer 1
# modifications has been completed, layer 2 are applied. This
# is how editors work and also the better way to handle 
# priority. It is not the same applying a greenscreen and then
# an effect than applying the effect first to the clip and then
# the greenscreen that wraps the whole video.


DEFAULT_SCENE_SIZE = (1920, 1080)
"""
A default scene size (FullHD) defined in pixels.
This size is the same as in moviepy because this
engine works also with pixels.
"""
MOVIEPY_SCENE_DEFAULT_SIZE = (1920, 1080)
"""
This is the default moviepy scene size we
use to simplify the way we calculate and 
use coordinates to position videos over
other videos when working with moviepy
video engine.
"""
MANIM_SCENE_DEFAULT_SIZE = (8, 14 + (2 / 9))
"""
The manim equivalent of an scene of 1920x1080 pixels,
adapted to be able to work with it when using the
manim engine.
"""


__all__ = [
    'MANIM_SCENE_DEFAULT_SIZE',
    'MOVIEPY_SCENE_DEFAULT_SIZE',
    'DEFAULT_SCENE_SIZE'
]


from yta_multimedia.audio.voice.transcription.stt.whisper import WhisperTranscriptor
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.file.checker import FileValidator
from typing import Union


class AudioTranscriptionModel(Enum):
    """
    The different models/engines available to transcribe
    audios.
    """
    WHISPER_TIMESTAMPED = 'whisper_timestamped'

class AudioTranscription:
    """
    Class that represent an audio transcription, which
    is an audio that has been transcripted by using a
    AI transcription model.
    """
    _text: str = None
    """
    The whole text that has been transcripted, plain,
    as it is, as a string.
    """
    _words: list[dict] = None
    """
    An array containing each word that has been 
    transcripted, including the 'start' and 'end' time
    moments in which they have been said in the audio.
    """

    @property
    def text(self):
        """
        The whole text that has been transcripted, plain,
        as it is, as a string.
        """
        if self._text is None:
            self._text = ' '.join(self.words['text'])

        return self._text

    @property
    def words(self):
        """
        An array containing each word that has been 
        transcripted, including the 'start' and 'end' time
        moments in which they have been said in the audio.
        """
        return self._words

    def __init__(self, words: list[dict]):
        EXPECTED_FIELDS = ['text', 'start', 'end']
        if any(not all(field in word for field in EXPECTED_FIELDS) for word in words):
            raise ValueError(f'All words provided must have these fields: "{', '.join(EXPECTED_FIELDS)}"')

        self._words = words

    @staticmethod
    def transcribe(audio_filename: str, initial_prompt: Union[str, None] = None):
        """
        Transcribe the provided 'audio_filename' with the 
        default model and using the 'initial_prompt' as a
        help (if provided).

        The initial prompt will be sent to the model to
        improve the way it transcribes the audio. This,
        if we know the text that must be transcripted
        but we want to know the timing, could be a whole
        string containing the expected output text.
        """
        if not FileValidator.file_is_audio_file(audio_filename):
            raise Exception('The provided "audio_filename" is not a valid audio file.')
        
        # We use this model by default
        words, _ = WhisperTranscriptor.transcribe_with_timestamps(audio_filename, initial_prompt = initial_prompt)

        return AudioTranscription(words)


from yta_multimedia.video.transition.objects import TransitionMode, TransitionMethod, VideoTransition, Transition
from yta_multimedia.video.position import Position
from moviepy.Clip import Clip


class AlphaTransition(Transition):
    """
    Transition between 2 videos that is made by applying
    an alpha video as a mask to make the first video
    transition to the second one.

    The alpha video must be, if possible, a pure black and
    white video (transparency is accepted) that goes from
    black to white, as black color will be opaque and white
    will be transparent letting the second video appear.
    """
    def __init__(self, video1: Clip, video2: Clip, mode: TransitionMode = TransitionMode.PLAYING, duration: float = 1.0, alpha_video: Clip = None):
        mode = TransitionMode.PLAYING if mode is None else TransitionMode.to_enum(mode)
        # TODO: Duration must be valid according to the
        # videos duration, so improve this condition
        duration = 1.0 if duration is None else duration

        if alpha_video is None:
            raise Exception('No "alpha_video" provided.')

        super().__init__(
            video1,
            VideoTransition(mode, duration, TransitionMethod.alpha, alpha_video = alpha_video),
            video2
        )

class SlideLeftTransition(Transition):
    """
    Transition between 2 videos that is made by making
    the first video disappear from one side of the
    scene while the second video is appearing at the
    same time.
    """
    def __init__(self, video1: Clip, video2: Clip, mode: TransitionMode = TransitionMode.PLAYING, duration: float = 1.0):
        mode = TransitionMode.PLAYING if mode is None else TransitionMode.to_enum(mode)
        # TODO: Duration must be valid according to the
        # videos duration, so improve this condition
        duration = 1.0 if duration is None else duration

        super().__init__(
            video1,
            VideoTransition(mode, duration, TransitionMethod.slide, side = Position.OUT_LEFT),
            video2
        )

class SlideRightTransition(Transition):
    """
    Transition between 2 videos that is made by making
    the first video disappear from one side of the
    scene while the second video is appearing at the
    same time.
    """
    def __init__(self, video1: Clip, video2: Clip, mode: TransitionMode = TransitionMode.PLAYING, duration: float = 1.0):
        mode = TransitionMode.PLAYING if mode is None else TransitionMode.to_enum(mode)
        # TODO: Duration must be valid according to the
        # videos duration, so improve this condition
        duration = 1.0 if duration is None else duration

        super().__init__(
            video1,
            VideoTransition(mode, duration, TransitionMethod.slide, side = Position.OUT_RIGHT),
            video2
        )

class SlideTopTransition(Transition):
    """
    Transition between 2 videos that is made by making
    the first video disappear from one side of the
    scene while the second video is appearing at the
    same time.
    """
    def __init__(self, video1: Clip, video2: Clip, mode: TransitionMode = TransitionMode.PLAYING, duration: float = 1.0):
        mode = TransitionMode.PLAYING if mode is None else TransitionMode.to_enum(mode)
        # TODO: Duration must be valid according to the
        # videos duration, so improve this condition
        duration = 1.0 if duration is None else duration

        super().__init__(
            video1,
            VideoTransition(mode, duration, TransitionMethod.slide, side = Position.OUT_TOP),
            video2
        )

class SlideBottomTransition(Transition):
    """
    Transition between 2 videos that is made by making
    the first video disappear from one side of the
    scene while the second video is appearing at the
    same time.
    """
    def __init__(self, video1: Clip, video2: Clip, mode: TransitionMode = TransitionMode.PLAYING, duration: float = 1.0):
        mode = TransitionMode.PLAYING if mode is None else TransitionMode.to_enum(mode)
        # TODO: Duration must be valid according to the
        # videos duration, so improve this condition
        duration = 1.0 if duration is None else duration

        super().__init__(
            video1,
            VideoTransition(mode, duration, TransitionMethod.slide, side = Position.OUT_BOTTOM),
            video2
        )


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.consts import MANIM_SCENE_DEFAULT_SIZE
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper
from yta_general_utils.programming.parameter_validator import NumberValidator
from typing import Union
from manim import *


class CinematicBarsManimAnimationWrapper(BaseManimAnimationWrapper):
    """
    Black bars that appear to make the scene cinematic.
    """
    duration: float = None

    def __init__(self, duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if not NumberValidator.is_positive_number(duration) or not NumberValidator.is_number_between(duration, 0, 100, do_include_lower_limit = False):
            exception_messages.append('The "duration" parameter provided is not a positive number between (0, 100] (zero is not valid).')

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        self.duration = duration
        super().__init__(CinematicBarsManimAnimationGenerator)

class CinematicBarsManimAnimationGenerator(BaseManimAnimation):
    """
    Black bars that appear to make the scene cinematic.
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        top_bar = Rectangle(BLACK, stroke_width = 0, fill_color = BLACK, fill_opacity = 1, height = MANIM_SCENE_DEFAULT_SIZE[1] * 0.15, width = MANIM_SCENE_DEFAULT_SIZE[0])
        bottom_bar = top_bar.copy()

        START_TOP_POSITION = (0, MANIM_SCENE_DEFAULT_SIZE[1] / 2 + (top_bar.height / 2), 0)
        STAY_TOP_POSITION = (0, MANIM_SCENE_DEFAULT_SIZE[1] / 2, 0)
        START_BOTTOM_POSITION = (0, -MANIM_SCENE_DEFAULT_SIZE[1] / 2 - (bottom_bar.height / 2), 0)
        STAY_BOTTOM_POSITION = (0, -MANIM_SCENE_DEFAULT_SIZE[1] / 2, 0)

        top_bar.move_to(START_TOP_POSITION)
        bottom_bar.move_to(START_BOTTOM_POSITION)

        self.add(top_bar)
        self.add(bottom_bar)

        # Appearing and dissapearing animation can be 2 seconds
        # longer as maximum (each one). If this happens, the 
        # stay will be longer
        animation_duration = 0.2 * self.parameters['duration']
        stay_duration = 0.6 * self.parameters['duration']
        if animation_duration > 2:
            stay_duration += (animation_duration - 2) * 2

        self.play(AnimationGroup([
            top_bar.animate.move_to(STAY_TOP_POSITION),
            bottom_bar.animate.move_to(STAY_BOTTOM_POSITION)
        ]), run_time = animation_duration, rate_func = linear)

        self.wait(stay_duration)

        self.play(AnimationGroup([
            top_bar.animate.move_to(START_TOP_POSITION),
            bottom_bar.animate.move_to(START_BOTTOM_POSITION)
        ]), run_time = animation_duration, rate_func = linear)







# class CinematicBarsManimAnimation(BaseManimAnimation):
#     """
#     Black bars that appear to make the scene cinematic.
#     """
#     def construct(self):
#         """
#         This method is called by manim when executed by shell and
#         will call the scene animation render method to be processed
#         and generated.
#         """
#         self.animate()

#     def generate(self, duration: float = 2, output_filename: str = 'output.mov'):
#         """
#         Checks and validates the provided parameters and generates
#         the manim animation if those parameters are valid.
#         """
#         # Check and validate all parameters
#         parameters = {}

#         if super().parameter_is_mandatory('duration', self.required_parameters) and not duration:
#             raise Exception('Field "duration" is mandatory. Aborting manim creation...')
#         if duration < 0 or duration > 100:
#             raise Exception('Field "duration" value is not valid. Must be between 0 and 100')
        
#         parameters['duration'] = duration

#         if not output_filename:
#             output_filename = 'output.mov'

#         # Generate the animation when parameters are valid
#         super().generate(parameters, output_filename = output_filename)

#         return output_filename

#     def animate(self):
#         top_bar = Rectangle(BLACK, stroke_width = 0, fill_color = BLACK, fill_opacity = 1, height = MANIM_SCENE_DEFAULT_SIZE[1] * 0.15, width = MANIM_SCENE_DEFAULT_SIZE[0])
#         bottom_bar = top_bar.copy()

#         START_TOP_POSITION = (0, MANIM_SCENE_DEFAULT_SIZE[1] / 2 + (top_bar.height / 2), 0)
#         STAY_TOP_POSITION = (0, MANIM_SCENE_DEFAULT_SIZE[1] / 2, 0)
#         START_BOTTOM_POSITION = (0, -MANIM_SCENE_DEFAULT_SIZE[1] / 2 - (bottom_bar.height / 2), 0)
#         STAY_BOTTOM_POSITION = (0, -MANIM_SCENE_DEFAULT_SIZE[1] / 2, 0)

#         top_bar.move_to(START_TOP_POSITION)
#         bottom_bar.move_to(START_BOTTOM_POSITION)

#         self.add(top_bar)
#         self.add(bottom_bar)

#         # Appearing and dissapearing animation can be 2 seconds
#         # longer as maximum (each one). If this happens, the 
#         # stay will be longer
#         animation_duration = 0.2 * self.parameters['duration']
#         stay_duration = 0.6 * self.parameters['duration']
#         if animation_duration > 2:
#             stay_duration += (animation_duration - 2) * 2

#         self.play(AnimationGroup([
#             top_bar.animate.move_to(STAY_TOP_POSITION),
#             bottom_bar.animate.move_to(STAY_BOTTOM_POSITION)
#         ]), run_time = animation_duration, rate_func = linear)

#         self.wait(stay_duration)

#         self.play(AnimationGroup([
#             top_bar.animate.move_to(START_TOP_POSITION),
#             bottom_bar.animate.move_to(START_BOTTOM_POSITION)
#         ]), run_time = animation_duration, rate_func = linear)